import re
import struct
import logging.config
from io import BytesIO

from .pdfafm import afm_loader
from .pdfcmap import cmap_manager, PDFCmap, PDFCMapInterpreter
from .pdfparser import PDFParser
from .pdfobject import PDFKeyWord, PDFName, PDFInteger
from .exception import PDFParserDoneException, PDFCMapNotFoundException
from .encoding import glyph_name_to_unicode, get_encoding

from .config import LOGGING

logging.config.dictConfig(LOGGING)
logger = logging.getLogger()


class PDFFont(object):
    def __init__(self, descriptor):
        self.descriptor = descriptor
        self.default_horizon_displacement = float(descriptor.get(b'MissingWidth', 0))
        self.horizon_displacements = {}
        self.default_vertical_displacement = 0
        self.vertical_displacements = {}
        # If the font in a source document uses a bold or italic style but there is no font data for that style,
        # the host operating system synthesizes the style. In this case, a com- ma and the style name
        # (one of Bold, Italic, or BoldItalic) are appended to the font name. For example, for a TrueType font
        # that is a bold variant of the New York font, the BaseFont value is written as /NewYork,Bold
        self.font_name = None
        if b'FontName' in descriptor:
            font_name = descriptor[b'FontName']
            if isinstance(font_name, PDFName):
                font_name = font_name.value
            if isinstance(font_name, str):
                font_name = font_name.encode()
            self.font_name = self.parse_font_name(font_name)
        self.flags = int(descriptor.get(b'Flags', 0))
        self._ascent = float(descriptor.get(b'Ascent', 0))
        self._descent = float(descriptor.get(b'Descent', 0))
        self.italic_angle = float(descriptor.get(b'ItalicAngle', 0))
        self.leading = float(descriptor.get(b'Leading', 0))
        self.bbox = list(descriptor.get(b'FontBBox', (0, 0, 0, 0)))
        self.horizon_scale = 0.001
        self.vertical_scale = 0.001

    @staticmethod
    def parse_font_name(name):
        # For a font subset, the PostScript name of the font—the value of the font’s BaseFont entry
        # and the font descriptor’s FontName entry—begins with a tag followed by a plus sign (+).
        # The tag consists of exactly six uppercase letters; the choice of letters is arbitrary,
        # but different subsets in the same PDF file must have different tags.
        name = re.sub(rb'^[A-Z]{6}\+', b'', name)
        try:
            name = name.decode()
        except UnicodeDecodeError:
            name = name.decode('gbk')
        return name

    def is_vertical(self):
        return False

    def is_multi_byte(self):
        return False

    def decode(self, bytes):
        return bytearray(bytes)

    @property
    def ascent(self):
        return self._ascent * self.vertical_scale

    @property
    def descent(self):
        return self._descent * self.vertical_scale

    @property
    def width(self):
        w = self.bbox[2] - self.bbox[0]
        if w == 0:
            w = - self.default_horizon_displacement
        return w * self.horizon_scale

    @property
    def height(self):
        h = self.bbox[3] - self.bbox[1]
        if h == 0:
            h = self._ascent - self._descent
        return h * self.vertical_scale

    def char_horizon_displacement(self, cid):
        if cid in self.horizon_displacements:
            return self.horizon_displacements[cid] * self.horizon_scale
        u = self.to_unicode(cid)
        if u in self.horizon_displacements:
            return self.horizon_displacements[u] * self.horizon_scale
        return self.default_horizon_displacement * self.horizon_scale

    def char_vertical_displacement(self, cid):
        return self.vertical_displacements.get(cid, self.default_vertical_displacement) * self.vertical_scale

    def to_unicode(self, cid):
        return None


class PDFSimpleFont(PDFFont):
    def __init__(self, descriptor, spec):
        PDFFont.__init__(self, descriptor)
        # Font encoding is specified either by a name of
        # built-in encoding or a dictionary that describes
        # the differences.
        encoding = spec.get(b'Encoding')
        if isinstance(encoding, dict):
            base_encoding = encoding[b'BaseEncoding'].value if b'BaseEncoding' in encoding else b'StandardEncoding'
            differences = [i.value for i in encoding.get(b'Differences', [])]
            self.cid_unicode = get_encoding(base_encoding, differences)
        else:
            self.cid_unicode = get_encoding(encoding.value if encoding else b'StandardEncoding')

        self.cmap = None
        if b'ToUnicode' in spec:
            stream = spec[b'ToUnicode']
            interpreter = PDFCMapInterpreter(BytesIO(stream.value))
            self.cmap = interpreter.cmap

    def to_unicode(self, cid):
        if self.cmap and cid in self.cmap.cid_unicode:
            return self.cmap.cid_unicode[cid]
        if cid in self.cid_unicode:
            return self.cid_unicode[cid]
        if cid == 10:
            return '\n'
        if cid == 13:
            return '\r'
        logging.warning(f'missing cid: {cid}')
        return ''


class PDFType1Font(PDFSimpleFont):
    def __init__(self, spec):
        self.basefont = self.parse_font_name(spec[b'BaseFont'].value) if b'BaseFont' in spec else None
        # For the standard 14 fonts, the entries FirstChar, LastChar, Widths, and FontDescriptor
        # must either all be present or all be absent. Ordinarily, they are absent;
        # specifying them enables a standard font to be overridden
        if b'FontDescriptor' in spec:
            descriptor = spec[b'FontDescriptor']
            first_char = int(spec[b'FirstChar'].value) if b'FirstChar' in spec else 0
            horizon_displacements = {
                first_char + i: w for (i, w) in enumerate(spec.get(b'Widths', [0] * 256))
            }
        else:
            descriptor = afm_loader.fonts[self.basefont]['descriptor']
            horizon_displacements = {k: v['WX'] for k, v in afm_loader.fonts[self.basefont]['metrics'].items()}

        if b'FontFile3' in descriptor:
            self.parse_font_file3(descriptor[b'FontFile3'])

        super().__init__(descriptor, spec)
        self.horizon_displacements = horizon_displacements

        if b'Encoding' not in spec and b'FontFile' in descriptor:
            # try to recover the missing encoding info from the font file.
            self.parse_font_file(descriptor[b'FontFile'])

    def parse_font_file(self, stream):
        # The length in bytes of the clear-text portion of the Type 1 font program
        length1 = int(stream[b'Length1'])
        data = stream.value[:length1]

        stack = []
        parser = PDFParser(BytesIO(data))
        while True:
            try:
                o = parser.next()
            except PDFParserDoneException:
                break
            if not isinstance(o, PDFKeyWord):
                stack.append(o)
            elif o.value == b'put':
                # dup 32 /space put
                # ...
                # dup 254 /bracerightbt put
                value = stack.pop()
                key = stack.pop()
                if isinstance(key, PDFInteger) and isinstance(value, PDFName):
                    self.cid_unicode[key.value] = glyph_name_to_unicode(value.value)

    def parse_font_file3(self, stream):
        pass


class PDFTrueTypeFont(PDFType1Font):
    pass


class PDFType3Font(PDFSimpleFont):
    def __init__(self, spec):
        if b'FontDescriptor' in spec:
            descriptor = spec[b'FontDescriptor']
        else:
            descriptor = {
                b'Ascent': 0,
                b'Descent': 0,
                b'FontBBox': spec[b'FontBBox']
            }
        super().__init__(descriptor, spec)

        first_char = int(spec[b'FirstChar'].value) if b'FirstChar' in spec else 0
        self.horizon_displacements = dict((i + first_char, w) for (i, w) in enumerate(spec.get(b'Widths', [0] * 256)))

        _, self._descent, _, self._ascent = self.bbox
        a, b, c, d, e, f = spec[b'FontMatrix']
        self.horizon_scale, self.vertical_scale = a + c, b + d


class PDFCIDFont(PDFFont):
    def __init__(self, spec):
        descriptor = spec.get(b'FontDescriptor') or {}
        super().__init__(descriptor)

        self.basefont = self.parse_font_name(spec[b'BaseFont'].value) if b'BaseFont' in spec else None
        cid_system_info = spec.get(b'CIDSystemInfo', {})
        cid_coding = b'-'.join([
            cid_system_info[b'Registry'].value if b'Registry' in cid_system_info else b'',
            cid_system_info[b'Ordering'].value if b'Ordering' in cid_system_info else b'',
        ])
        encoding = spec[b'Encoding'].value if b'Encoding' in spec else None

        self.cmap = cmap_manager.get_cmap(encoding) or PDFCmap()

        self.unicode_map = PDFCmap()
        if b'ToUnicode' in spec:
            stream = spec[b'ToUnicode']
            interpreter = PDFCMapInterpreter(BytesIO(stream.value))
            self.unicode_map = interpreter.cmap
        elif cid_coding in (b'Adobe-Identity', b'Adobe-UCS'):
            if b'FontFile2' in descriptor:
                self.parse_font_file(descriptor[b'FontFile2'].value)
        else:
            self.unicode_map = cmap_manager.get_unicode_map(cid_coding, 'v' if self.cmap.is_vertical() else 'h')

        if self.is_vertical():
            widths = W2(spec.get(b'W2', []))
            vy, w = spec.get(b'DW2', [880, -1000])
            self.default_vertical_displacement = vy
            self.horizon_displacements = dict((cid, w) for (cid, (w, _)) in widths.items())
            self.default_horizon_displacement = w
            self.vertical_displacements = dict((cid, vy) for (cid, (_, (vx, vy))) in widths.items())
        else:
            self.horizon_displacements = W(spec.get(b'W', {}))
            self.default_horizon_displacement = spec.get(b'DW', 1000)

    def parse_font_file(self, font_file):
        # create unicode map
        self.unicode_map = PDFCmap()

        fp = BytesIO(font_file)
        font_type = fp.read(4)

        tables = {}
        try:
            table_number, _, _, _ = struct.unpack('>HHHH', fp.read(8))
            for _ in range(table_number):
                name, _, offset, length = struct.unpack('>4sLLL', fp.read(16))
                tables[name] = (offset, length)
        except struct.error:
            # Do not fail if there are not enough bytes to read. Even for
            # corrupted PDFs we would like to get as much information as
            # possible, so continue.
            pass

        if b'cmap' not in tables:
            return
        base_offset, length = tables[b'cmap']
        fp.seek(base_offset)
        version, sub_table_number = struct.unpack('>HH', fp.read(4))
        sub_tables = [base_offset + struct.unpack('>HHL', fp.read(8)) for _ in range(sub_table_number)]
        char2gid = {}
        # Only supports sub_tables type 0, 2 and 4.
        for _, _, sub_table_offset in sub_tables:
            fp.seek(sub_table_offset)
            fmt_type, fmt_len, fmt_lang = struct.unpack('>HHH', fp.read(6))
            if fmt_type == 0:
                char2gid.update(enumerate(struct.unpack('>256B', fp.read(256))))
            elif fmt_type == 2:
                sub_header_keys = struct.unpack('>256H', fp.read(512))
                first_bytes = [0] * 8192
                for (i, k) in enumerate(sub_header_keys):
                    first_bytes[k//8] = i
                nhdrs = max(sub_header_keys)//8 + 1
                hdrs = []
                for i in range(nhdrs):
                    (firstcode, entcount, delta, offset) = struct.unpack('>HHhH', fp.read(8))
                    hdrs.append((i, firstcode, entcount, delta, fp.tell()-2+offset))
                for (i, firstcode, entcount, delta, pos) in hdrs:
                    if not entcount:
                        continue
                    first = firstcode + (first_bytes[i] << 8)
                    fp.seek(pos)
                    for c in range(entcount):
                        gid = struct.unpack('>H', fp.read(2))
                        if gid:
                            gid += delta
                        char2gid[first+c] = gid
            elif fmt_type == 4:
                (segcount, _1, _2, _3) = struct.unpack('>HHHH', fp.read(8))
                segcount //= 2
                ecs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
                fp.read(2)
                scs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
                idds = struct.unpack('>%dh' % segcount, fp.read(2*segcount))
                pos = fp.tell()
                idrs = struct.unpack('>%dH' % segcount, fp.read(2*segcount))
                for (ec, sc, idd, idr) in zip(ecs, scs, idds, idrs):
                    if idr:
                        fp.seek(pos+idr)
                        for c in range(sc, ec+1):
                            char2gid[c] = (struct.unpack('>H', fp.read(2))[0] + idd) & 0xffff
                    else:
                        for c in range(sc, ec+1):
                            char2gid[c] = (c + idd) & 0xffff
            else:
                assert False, str(('Unhandled', fmt_type))
        for char, gid in char2gid.items():
            self.unicode_map.add_cid_unicode(gid, char)

    def is_vertical(self):
        return self.cmap.is_vertical()

    def is_multi_byte(self):
        return True

    def decode(self, bytes):
        return self.cmap.decode(bytes)

    def to_unicode(self, cid):
        return self.unicode_map.get_unicode(cid)


def W(ws):
    """
    c [w1 w2 ... wn]
    cfirst clast w
    """
    widths = {}
    while ws:
        t = ws.pop()
        if isinstance(t, list):
            c = ws.pop()
            for i, w in enumerate(t):
                widths[c + i] = w
        else:
            c2 = ws.pop()
            c1 = ws.pop()
            for c in range(c1, c2 + 1):
                widths[c] = t
    return widths


def W2(ws):
    """
    c [w11y v1x v1y w12y v2x v2y ...]
    cfirst clast w11y v1x v1y
    """
    widths = {}
    while ws:
        t = ws.pop()
        if isinstance(t, list):
            c = ws.pop()
            i = 0
            while t:
                vy = t.pop()
                vx = t.pop()
                w = t.pop()
                widths[c + i] = (w, (vx, vy))
                i += 1
        else:
            vy = t
            vx = ws.pop()
            w = ws.pop()
            c2 = ws.pop()
            c1 = ws.pop()
            for c in range(c1, c2 + 1):
                widths[c] = (w, (vx, vy))
    return widths



import struct


def ascii_85_decode(data):
    decoded = b''
    n = b = 0
    for c in data:
        if b'!' <= c <= b'u':
            n += 1
            b = b * 85 + (ord(c) - 33)
            if n == 5:
                decoded += struct.pack('>L', b)
                n = b = 0
        elif c == b'z':
            decoded += b'\0\0\0\0'
        elif c == b'~':
            if n:
                for _ in range(5-n):
                    b = b * 85 + 84
                decoded += struct.pack('>L', b)[:n-1]
            break
    return decoded




from .ascii_85_decoder import ascii_85_decode
from .ascii_hex_decoder import ascii_hex_decode
from .flate_decoder import flate_decode
from .lzw_decoder import lzw_decode
from .run_length_decoder import run_length_decode


DECODER_MAP = {
    b'FlateDecode': flate_decode,
    b'Fl': flate_decode,

    b'LZWDecode': lzw_decode,
    b'LZW': lzw_decode,

    b'ASCII85Decode': ascii_85_decode,
    b'A85': ascii_85_decode,

    b'ASCIIHexDecode': ascii_hex_decode,
    b'AHx': ascii_hex_decode,

    b'RunLengthDecode': run_length_decode,
    b'RL': run_length_decode,

    b'CCITTFaxDecode': None,
    b'CCF': None,

    b'DCTDecode': None,
    b'DCT': None,

    b'JBIG2Decode': None,
    b'JBIG2': None,
}


__all__ = [
    'DECODER_MAP',

    'ascii_85_decode',
    'ascii_hex_decode',
    'flate_decode',
    'lzw_decode',
    'run_length_decode',
]



import struct


def pack(code, length):
    return struct.pack('>L', code)[-length:]


def unpack(s):
    length = len(s)
    if length == 1:
        return ord(s)
    elif length == 2:
        return struct.unpack('>H', s)[0]
    elif length == 3:
        return struct.unpack('>L', b'\x00'+s)[0]
    elif length == 4:
        return struct.unpack('>L', s)[0]
    elif length == 8:
        return struct.unpack('>Q', s)[0]




import regex as re

PATTERN = re.compile(
    br'('
    br'(?P<pair>[0-9a-f]{2})'
    br'|\s'
    br')*'
    br'(?P<single>[0-9a-f])?',
    re.IGNORECASE
)


def ascii_hex_decode(data):
    decoded = b''
    match = PATTERN.fullmatch(data)
    if match:
        d = match.capturesdict()
        if d.get('pair'):
            decoded += bytes(int(pair, 16) for pair in d['pair'])
        if d.get('single'):
            decoded += bytes(int(single + b'0', 16) for single in d['single'])
    return decoded



import re
import logging.config

from .config import LOGGING
from .exception import PDFLexerException, PDFLexerEOLException
from .pdfcharset import (
    EOL,
    NON_EOL,
    WHITE_SPACE,
    NON_SPACE,
    REGULAR,
    NON_REGULAR,
    NUMBER,
    NON_NUMBER,
    STRING,
    NON_STRING,
    HEX_STRING,
    NON_HEX_STRING,
    NON_KEYWORD,
)

logging.config.dictConfig(LOGGING)
logger = logging.getLogger()


class Token(object):
    def __init__(self, name, text):
        self.name = name
        self.text = text

    def __eq__(self, other):
        return self.name == other.name and self.text == other.text

    def __repr__(self):
        return f'Token("{self.name}", {self.text})'


class PDFLexer(object):

    def __init__(self, fp, buffer_size=4096):
        self.fp = fp
        self.fp.seek(0)

        self._buffer = b''
        self._buffer_size = buffer_size
        self._buffer_position = self.fp.tell()

    def tell(self):
        return self._buffer_position

    def seek(self, position, whence=0):
        self.fp.seek(position, whence)
        self._buffer = b''
        self._buffer_position = self.fp.tell()

    def reversed_lines(self):
        buffer = b''
        self.seek(0, 2)
        cur_pos = self.fp.tell()
        while cur_pos > 0:
            next_pos = max(0, cur_pos - self._buffer_size)
            self.seek(next_pos)
            s = self.fp.read(cur_pos - next_pos)
            if not s:
                break
            buffer = s + buffer

            while EOL.search(buffer):
                lines = re.split(rb'[\r\n]+', buffer)
                buffer = lines.pop(0)
                for line in reversed(lines):
                    yield line
            cur_pos = next_pos

        if buffer:
            yield buffer

    def next_line(self):
        s = self.read_until(EOL)
        self.read_until(NON_EOL)
        return s

    @property
    def buffer(self):
        if not self._buffer:
            self._buffer_position = self.fp.tell()
            self._buffer = self.fp.read(self._buffer_size)
        if not self._buffer:
            raise EOFError()
        return self._buffer

    def look_some(self, num):
        position = self.fp.tell()
        buffer = self._buffer
        buffer_position = self._buffer_position
        s = self.read_some(num)
        self.seek(position)
        self._buffer = buffer
        self._buffer_position = buffer_position
        return s

    def read_some(self, num):
        tmp = []
        while num:
            if len(self.buffer) <= num:
                num -= len(self.buffer)
                tmp.append(self.buffer)
                self._buffer = b''
            else:
                tmp.append(self.buffer[:num])
                self._buffer = self.buffer[num:]
                self._buffer_position += num
                break
        return b''.join(tmp)

    def read_until(self, stop):
        buffer = b''
        match = False
        buffer_length = 0
        while not match:
            buffer += self.buffer
            match = stop.search(buffer)
            buffer_length = len(self._buffer)
            self._buffer = b''
        rest_buffer_length = len(buffer) - match.start()
        self._buffer_position += buffer_length - rest_buffer_length
        self._buffer = buffer[match.start():]
        return buffer[:match.start()]

    def number_token(self):
        s = self.read_until(NUMBER)
        s += self.read_until(NON_NUMBER)
        if b'.' not in s and self.buffer[0:1] == b'.':
            s += self.read_some(1)
            s += self.read_until(NON_NUMBER)
        return Token('real', s) if b'.' in s else Token('integer', s)

    def literal_string_token(self):
        escapes = {
            b'n': b'\n',
            b'r': b'\r',
            b't': b'\t',
            b'b': b'\b',
            b'f': b'\f',
        }

        self.read_some(1)
        s = b''
        level = 0
        while self.buffer:
            if self.buffer[0:1] == b'\\':
                self.read_some(1)
                t = self.read_some(1)
                if t in escapes:
                    s += escapes[t]
                elif t in b'\r\n':
                    while self.buffer[0:1] in b'\r\n':
                        self.read_some(1)
                elif t in b'01234567':
                    while len(t) < 3 and self.buffer[0:1] in b'01234567':
                        t += self.read_some(1)
                    s += bytes([int(t, 8)])
                else:
                    s += t
            elif self.buffer[0:1] == b'(':
                self.read_some(1)
                level += 1
            elif self.buffer[0:1] == b')':
                if level == 0:
                    break
                self.read_some(1)
                level -= 1
            else:
                self.read_until(STRING)
                s += self.read_until(NON_STRING)
        else:
            raise PDFLexerException(f'Missing Parentheses: {s}')
        self.read_some(1)
        return Token('string', s)

    def hex_string_token(self):
        self.read_some(1)
        s = self.read_until(NON_HEX_STRING)
        s = WHITE_SPACE.sub(b'', s)

        tmp = []
        for i in range(len(s)):
            if not i % 2:
                tmp.append(bytes([int(s[i: i + 2], 16)]))
        s = b''.join(tmp)
        self.read_some(1)
        return Token('string', s)

    def array_token(self):
        tokens = []
        self.read_some(1)
        while True:
            self.read_until(NON_SPACE)
            if self.buffer[0:1] == b']':
                break
            tokens.append(self.next_token())
        self.read_some(1)
        return tokens

    def array_begin_token(self):
        s = self.read_some(1)
        return Token('array_begin', s)

    def array_end_token(self):
        s = self.read_some(1)
        return Token('array_end', s)

    def dict_begin_token(self):
        s = self.read_some(2)
        return Token('dict_begin', s)

    def dict_end_token(self):
        s = self.read_some(2)
        return Token('dict_end', s)

    def name_token(self):
        self.read_until(REGULAR)
        s = self.read_until(NON_REGULAR)
        for match in re.finditer(rb'#(?P<hex>[0-9a-f]{2})', s, flags=re.I):
            s = s.replace(match.group(), bytes([int(match.groupdict()['hex'], 16)]))
        return Token('name', s)

    def keyword_token(self):
        s = self.read_until(NON_KEYWORD)
        return Token('keyword', s)

    def next_token(self):
        try:
            self.read_until(NON_SPACE)
            c = self.look_some(1)
        except EOFError:
            raise PDFLexerEOLException()

        if c == b'%':
            self.read_until(EOL)
            token = self.next_token()
        elif c in b'-+.0123456789':
            return self.number_token()
        elif c == b'(':
            token = self.literal_string_token()
        elif c == b'<':
            if self.look_some(2) == b'<<':
                token = self.dict_begin_token()
            else:
                token = self.hex_string_token()
        elif c == b'>':
            if self.look_some(2) == b'>>':
                token = self.dict_end_token()
            else:
                raise PDFLexerException(f'Unexpected char: >')
        elif c == b'[':
            token = self.array_begin_token()
        elif c == b']':
            return self.array_end_token()
        elif c == b'/':
            token = self.name_token()
        else:
            token = self.keyword_token()
        # logger.debug(f"token: {token}")
        return token



import re
import logging.config

from .config import LOGGING
from .exception import PDFParserException, PDFParserDoneException
from .pdflexer import PDFLexer
from .pdfobject import PDF_OBJECT_DICT, PDFObj, PDFIndirectRef, PDFStream, PDFXref
from .utils.struct import unpack

logging.config.dictConfig(LOGGING)
logger = logging.getLogger()


class PDFParser(object):

    def __init__(self, fp):
        self.lexer = PDFLexer(fp)

        self._buffer = []
        self._buffer_positions = []

    @property
    def buffer(self):
        if len(self._buffer) < 3:
            for i in range(3 - len(self._buffer)):
                position = self.lexer.tell()
                try:
                    token = self.lexer.next_token()
                except:
                    break
                self._buffer_positions.append(position)
                self._buffer.append(token)

        return self._buffer

    def look_some(self, num):
        return self.buffer[:num]

    def read_some(self, num):
        tmp = []
        while num:
            if len(self.buffer) <= num:
                num -= len(self.buffer)
                tmp.extend(self.buffer)
                self._buffer = []
                self._buffer_positions = []
            else:
                tmp.extend(self.buffer[:num])
                self._buffer = self.buffer[num:]
                self._buffer_positions = self._buffer_positions[num:]
                break
        # logger.debug(f"tokens: {tmp}")
        return tmp

    def read_until(self, stop):
        if self._buffer_positions:
            self.seek(self._buffer_positions[0])
        self._buffer = []
        self._buffer_positions = []
        return self.lexer.read_until(stop)

    def next_line(self):
        if self._buffer_positions:
            self.seek(self._buffer_positions[0])
        self._buffer = []
        self._buffer_positions = []
        return self.lexer.next_line()

    def seek(self, position, whence=0):
        self._buffer = []
        self._buffer_positions = []
        self.lexer.seek(position, whence)

    def parse_object(self):
        token = self.read_some(1)[0]
        return PDF_OBJECT_DICT[token.name](token.text)

    def parse_array(self):
        self.read_some(1)
        o = []
        while True:
            token = self.look_some(1)[0]
            if token.name == 'array_end':
                self.read_some(1)
                break
            o.append(self.next())
        return o

    def parse_dict(self):
        self.read_some(1)
        o = {}
        k = None
        while True:
            token = self.look_some(1)[0]
            if token.name == 'dict_end':
                self.read_some(1)
                break
            if k is None:
                self.read_some(1)
                if token.name != 'name':
                    raise PDFParserException(f'Wrong Dict Key Type: {token}')
                k = token.text
            else:
                o[k] = self.next()
                k = None
        return o

    def parse_indirect_ref(self):
        tokens = self.read_some(3)
        return PDFIndirectRef(int(tokens[0].text), int(tokens[1].text))

    def parse_obj(self):
        tokens = self.read_some(3)
        n = self.next()
        t = self.look_some(1)[0]
        if t.name == 'keyword' and t.text == b'stream':
            n = self.parse_stream(n)
        endobj = self.read_some(1)[0]
        if endobj.name != 'keyword' or endobj.text != b'endobj':
            raise PDFParserException(f'Bad Obj: {endobj}')
        if isinstance(n, PDFStream):
            if b'Type' in n.dict and n.dict[b'Type'].value == b'XRef':
                return self.parse_xref_stream(n)
        return PDFObj(int(tokens[0].text), int(tokens[1].text), n)

    def parse_stream(self, d):
        self.read_some(1)
        stream = self.read_until(re.compile(rb'endstream'))
        endstream = self.read_some(1)[0]
        if endstream.name != 'keyword' or endstream.text != b'endstream':
            raise PDFParserException(f'Bad Stream: {endstream}')
        return PDFStream(d, stream)

    def parse_xref(self):
        self.read_some(1)
        xref = PDFXref()
        while True:
            token = self.look_some(1)[0]
            if token.text == b'trailer':
                self.read_some(1)
                xref.trailer = self.next()
                break
            line = self.next_line().strip()
            if not line:
                continue
            start, num = map(int, line.split(b' '))
            for obj_id in range(start, start + num):
                line = self.next_line().strip()
                pos, gen_no, use = line.split(b' ')
                if use != b'n':
                    continue
                xref.offsets[obj_id] = (None, int(pos), int(gen_no))
        return xref

    @staticmethod
    def parse_xref_stream(stream):
        xref = PDFXref()
        xref.trailer = stream.dict

        size = xref.trailer[b'Size']
        index = xref.trailer.get(b'Index') or [0, size]
        w = xref.trailer[b'W']
        entry_length = sum(w)

        offset = 0
        while index:
            begin, number = index[:2]
            index = index[2:]

            for i in range(number):
                entry = stream.value[offset: offset + entry_length]
                entry_type = unpack(entry[:w[0]]) if entry[:w[0]] else 1
                if entry_type == 1:
                    pos = unpack(entry[w[0]: w[0] + w[1]])
                    gen_no = unpack(entry[w[0] + w[1]:]) if entry[w[0] + w[1]:] else 0
                    xref.offsets[begin + i] = (None, int(pos), gen_no)
                elif entry_type == 2:
                    stream_object_number = unpack(entry[w[0]: w[0] + w[1]])
                    obj_pos = unpack(entry[w[0] + w[1]:]) if entry[w[0] + w[1]:] else 0
                    xref.offsets[begin + i] = (int(stream_object_number), int(obj_pos), 0)
                offset += entry_length
        return xref

    def next(self):
        tokens = self.look_some(3)
        if not tokens:
            raise PDFParserDoneException

        if len(tokens) == 3 and tokens[2].name == 'keyword':
            if tokens[2].text == b'R':
                return self.parse_indirect_ref()
            elif tokens[2].text == b'obj':
                return self.parse_obj()

        token = tokens[0]
        if token.name == 'array_begin':
            return self.parse_array()
        elif token.name == 'dict_begin':
            return self.parse_dict()
        elif token.name == 'keyword' and token.text == b'xref':
            return self.parse_xref()
        else:
            return self.parse_object()



import os
import codecs
import struct
from collections import defaultdict
from copy import deepcopy
import logging.config

from .utils.struct import pack, unpack
from .pdfparser import PDFParser
from .exception import PDFParserDoneException, PDFCMapNotFoundException
from .pdfobject import PDFKeyWord, PDFName
from .encoding import glyph_name_to_unicode


from .config import LOGGING

logging.config.dictConfig(LOGGING)
logger = logging.getLogger()


class PDFCmap(object):
    def __init__(self, attrs=None):
        self.attrs = attrs or {}
        self.code_cid = {}
        self.cid_unicode = {}

    def is_vertical(self):
        return self.attrs.get(b'WMode', 0) != 0

    def set_attr(self, k, v):
        self.attrs[k] = v
        if k == b'CMapName':
            cmap_manager.set_cmap(v.value, self)

    def use_cmap(self, cmap_name):
        cmap = cmap_manager.get_cmap(cmap_name)
        if cmap:
            code_cid = deepcopy(cmap.code_cid)
            self.code_cid.update(code_cid)

    def add_code_cid(self, code, cid, force=True):
        d = self.code_cid
        for c in code[:-1]:
            if c in d:
                d = d[c]
            else:
                t = {}
                d[c] = t
                d = t
        c = code[-1]
        if force or c not in d:
            d[c] = cid

    def add_cid_unicode(self, cid, code):
        if isinstance(code, PDFName):
            # Interpret as an Adobe glyph name.
            self.cid_unicode[cid] = glyph_name_to_unicode(code.value)
        elif isinstance(code, bytes):
            # Interpret as UTF-16BE.
            self.cid_unicode[cid] = code.decode('UTF-16BE', 'ignore')
        elif isinstance(code, int):
            self.cid_unicode[cid] = chr(code)
        else:
            self.cid_unicode[cid] = code

    def get_unicode(self, cid):
        if cid in self.cid_unicode:
            return self.cid_unicode[cid]
        return ''

    def decode(self, code):
        d = self.code_cid
        for c in code:
            if c in d:
                d = d[c]
                if isinstance(d, int):
                    yield d
                    d = self.code_cid
            else:
                d = self.code_cid


class PDFIdentityCmap(PDFCmap):
    def decode(self, code):
        n = len(code) // 2
        return struct.unpack(f'>{n}H', code) if n else ()


class PDFCMapInterpreter(object):

    def __init__(self, fp):
        self.parser = PDFParser(fp)
        self.cmap = PDFCmap()
        self.stack = []
        self.interpret()

    def interpret(self):
        while True:
            try:
                o = self.parser.next()
            except PDFParserDoneException:
                break
            if not isinstance(o, PDFKeyWord):
                self.stack.append(o)
            elif o.value == b'begincmap':
                self.stack = []
            elif o.value == b'endcmap':
                pass
            elif o.value == b'def':
                v = self.stack.pop()
                k = self.stack.pop().value
                self.cmap.set_attr(k, v)
            elif o.value == b'usecmap':
                cmap_name = self.stack.pop()
                self.cmap.use_cmap(cmap_name)
            elif o.value == b'begincodespacerange':
                self.stack = []
            elif o.value == b'endcodespacerange':
                self.stack = []
            elif o.value == b'begincidrange':
                self.stack = []
            elif o.value == b'endcidrange':
                while self.stack:
                    cid = self.stack.pop().value
                    end = self.stack.pop().value
                    begin = self.stack.pop().value

                    length = len(begin[-4:])
                    b1, b2 = begin[:-4], unpack(begin[-4:])
                    e1, e2 = end[:-4], unpack(end[-4:])
                    for i in range(e2 - b2 + 1):
                        x = b1 + pack(b2 + i, length)
                        self.cmap.add_code_cid(x, cid + i)
            elif o.value == b'begincidchar':
                self.stack = []
            elif o.value == b'endcidchar':
                while self.stack:
                    cid = self.stack.pop().value
                    code = unpack(self.stack.pop().value)
                    self.cmap.add_code_cid(code, cid)
            elif o.value == b'beginbfrange':
                self.stack = []
            elif o.value == b'endbfrange':
                while self.stack:
                    code = self.stack.pop()
                    end = unpack(self.stack.pop().value)
                    begin = unpack(self.stack.pop().value)
                    if isinstance(code, list):
                        for i in range(end - begin + 1):
                            self.cmap.add_cid_unicode(begin + i, code[i].value)
                    else:
                        code = code.value
                        length = len(code[-4:])
                        c1, c2 = code[:-4], unpack(code[-4:])
                        for i in range(end - begin + 1):
                            x = c1 + pack(c2 + i, length)
                            self.cmap.add_cid_unicode(begin + i, x)
            elif o.value == b'beginbfchar':
                self.stack = []
            elif o.value == b'endbfchar':
                while self.stack:
                    code = self.stack.pop().value
                    cid = unpack(self.stack.pop().value)
                    self.cmap.add_cid_unicode(cid, code)
            elif o.value == b'beginnotdefrange':
                self.stack = []
            elif o.value == b'endnotdefrange':
                self.stack = []


class CmapLoader(object):
    CODEC_DICT = {
        'B5': 'cp950',
        'GBK-EUC': 'cp936',
        'RKSJ': 'cp932',
        'EUC': 'euc-jp',
        'KSC-EUC': 'euc-kr',
        'KSC-Johab': 'johab',
        'KSCms-UHC': 'cp949',
        'UniCNS-UTF8': 'utf-8',
        'UniGB-UTF8': 'utf-8',
        'UniJIS-UTF8': 'utf-8',
        'UniKS-UTF8': 'utf-8',
    }

    def __init__(self):
        self.cmap_dict = {}
        self.unicode_map_dict = {}

    def load_cmap(self):
        cmap_path = os.path.join(os.path.dirname(__file__), 'cmap')
        for d in os.listdir(cmap_path):
            cid2code = os.path.join(cmap_path, d, 'cid2code.txt')
            if not os.path.isfile(cid2code):
                continue
            unicode_map_name = '-'.join(d.split('-')[:-1]).encode()
            self.unicode_map_dict[unicode_map_name] = {'v': PDFCmap(attrs={b'WMode': 1}), 'h': PDFCmap()}
            headers = []
            with open(cid2code) as f:
                for line in f:
                    line, _, _ = line.strip().partition('#')
                    if not line:
                        continue
                    row = line.split('\t')
                    if row[0] == 'CID':
                        headers = row
                        continue
                    if not headers:
                        continue
                    cid = int(row[0])
                    unicode_counter_h = defaultdict(int)
                    unicode_counter_v = defaultdict(int)
                    for header, value in zip(headers[1:], row[1:]):
                        if value == '*':
                            continue
                        code_h = []
                        code_v = []
                        for code in value.split(','):
                            vertical = code.endswith('v')
                            if vertical:
                                code = code[:-1]
                            try:
                                code = codecs.decode(code, 'hex_codec')
                            except:
                                code = chr(int(code, 16))
                            try:
                                u = code.decode(self.CODEC_DICT[header], 'strict')
                            except:
                                u = ''
                            if vertical:
                                code_v.append(code)
                                if len(u) == 1:
                                    unicode_counter_v[u] += 1
                            else:
                                code_h.append(code)
                                if len(u) == 1:
                                    unicode_counter_h[u] += 1

                        cmap_h, cmap_v = self.get_cmap(header)
                        if code_v:
                            for code in code_v:
                                cmap_v.add_code_cid(code, cid)
                            for code in code_h:
                                cmap_h.add_code_cid(code, cid)
                        else:
                            for code in code_h:
                                cmap_v.add_code_cid(code, cid, force=False)
                                cmap_h.add_code_cid(code, cid)
                    if unicode_counter_h:
                        self.unicode_map_dict[unicode_map_name]['h'].add_cid_unicode(
                            cid,
                            self.pick_unicode(unicode_counter_h)
                        )
                    if unicode_counter_v or unicode_counter_h:
                        self.unicode_map_dict[unicode_map_name]['v'].add_cid_unicode(
                            cid,
                            self.pick_unicode(unicode_counter_v or unicode_counter_h)
                        )

    @staticmethod
    def pick_unicode(counter):
        chars = sorted(counter.items(), key=(lambda x: (x[1], -ord(x[0]))), reverse=True)
        return chars[0][0]

    def get_cmap(self, name):
        name = name.encode()
        if name.endswith(b'-H'):
            name_h, name_v = name, None
        elif name == b'H':
            name_h, name_v = name, b'V'
        else:
            name_h, name_v = name + b'-H', name + b'-V'

        if name_h not in self.cmap_dict:
            self.cmap_dict[name_h] = PDFCmap()
        if name_v:
            if name_v not in self.cmap_dict:
                self.cmap_dict[name_v] = PDFCmap(attrs={b'WMode': 1})
        return self.cmap_dict[name_h], self.cmap_dict.get(name_v)


class CmapManager(object):
    CMAP_DICT = {}

    def __init__(self):
        self.cmap_loader = CmapLoader()
        self.cmap_dict = {}
        self.init()

    def init(self):
        self.cmap_dict = deepcopy(self.CMAP_DICT)
        self.cmap_loader.load_cmap()

    def set_cmap(self, cmap_name, cmap):
        self.cmap_dict[cmap_name] = cmap

    def get_cmap(self, cmap_name):
        if cmap_name == b'Identity-H':
            return PDFIdentityCmap(attrs={b'WMode': 0})
        if cmap_name == b'Identity-V':
            return PDFIdentityCmap(attrs={b'WMode': 1})
        if cmap_name in self.cmap_dict:
            return self.cmap_dict[cmap_name]
        if cmap_name in self.cmap_loader.cmap_dict:
            return self.cmap_loader.cmap_dict[cmap_name]

        raise PDFCMapNotFoundException(f'cmap not found: {cmap_name}')

    def get_unicode_map(self, unicode_map_name, wmode='h'):
        if unicode_map_name in self.cmap_loader.unicode_map_dict:
            return self.cmap_loader.unicode_map_dict[unicode_map_name][wmode]

        raise PDFCMapNotFoundException(f'unicode_map not found: {unicode_map_name}')


cmap_manager = CmapManager()





class PageReconstructor(object):
    THRESHOLD = 3

    def __init__(self, page):
        self.page = page
        self.horizontal_lines = {}
        self.vertical_lines = {}
        self.vertical_lines_groups = {}

        self.chars_box = [None, None, None, None]
        self.chars_ids = {}
        self.chars = {}
        self.ys = []
        self.strings = {}

        self.vertical_borders = {}
        self.horizontal_borders = {}
        self.tables = []

    def near_enough(self, a, b):
        return abs(a - b) < self.THRESHOLD

    def reconstruct(self):
        self.read_lines()
        self.horizontal_lines = self.join_lines(self.merge_lines(self.horizontal_lines))
        self.vertical_lines_groups = {
            k: self.join_lines(self.merge_lines(v)) for k, v in self.group_lines(self.vertical_lines).items()
        }
        self.vertical_lines = {}
        for k, vertical_lines in self.vertical_lines_groups.items():
            for x, lines in vertical_lines.items():
                if x not in self.vertical_lines:
                    self.vertical_lines[x] = []
                self.vertical_lines[x].extend(lines)
        self.read_chars()
        self.read_vertical_borders()
        self.read_tables()
        return sorted(list(self.strings.values()) + self.tables, key=lambda x: x['bounds'][1], reverse=True)

    def read_lines(self):
        for item in self.page['data']:
            if 'rectangle' in item or 'line' in item:
                x0, y0, x1, y1 = item.get('rectangle') or item.get('line')
                x0, x1 = min(x0, x1), max(x0, x1)
                y0, y1 = min(y0, y1), max(y0, y1)
                # 竖线
                if self.near_enough(x0, x1):
                    if x0 not in self.vertical_lines:
                        self.vertical_lines[x0] = []
                    self.vertical_lines[x0].append((y0, y1))
                # 横线
                elif self.near_enough(y0, y1):
                    if y0 not in self.horizontal_lines:
                        self.horizontal_lines[y0] = []
                    self.horizontal_lines[y0].append((x0, x1))

    def merge_lines(self, lines):
        merged_lines = {}
        ps = []
        for p in sorted(lines.keys()):
            if not ps or not self.near_enough(ps[-1], p):
                ps.append(p)
                merged_lines[p] = lines[p]
            else:
                if isinstance(lines[p], list):
                    merged_lines[ps[-1]].extend(lines[p])
                elif isinstance(lines[p], dict):
                    merged_lines[ps[-1]].update(lines[p])
        return merged_lines

    def join_lines(self, lines):
        joint_lines = {}
        for p in lines.keys():
            tmp_lines = []
            begin, end = None, None
            for b, e in sorted(lines[p]):
                if begin is None:
                    begin, end = b, e
                else:
                    if b <= end or self.near_enough(b, end):
                        end = max(e, end)
                    else:
                        tmp_lines.append((begin, end))
                        begin, end = b, e
            if begin is not None:
                tmp_lines.append((begin, end))
            joint_lines[p] = tmp_lines
        return joint_lines

    def group_lines(self, lines):
        lines_group = {}
        lines_groups = {}
        begin, end = None, None
        for p, b, e in sorted([(p, b, e) for p, v in lines.items() for b, e in v], key=lambda x: x[1]):
            if end is None:
                begin, end = b, e
            if b <= end or self.near_enough(b, end):
                end = max(e, end)
                if p not in lines_group:
                    lines_group[p] = []
                lines_group[p].append((b, e))
            else:
                lines_groups[(begin, end)] = lines_group
                begin, end = b, e
                lines_group = {}
        if lines_group:
            lines_groups[(begin, end)] = lines_group
        return lines_groups

    def read_chars(self):
        chars = {}
        heights = {}
        widths = {}
        for i, item in enumerate(self.page['data']):
            if 'char' not in item:
                continue
            self.chars_ids[i] = item
            if self.chars_box[0] is None or item['x0'] < self.chars_box[0]:
                self.chars_box[0] = item['x0']
            if self.chars_box[1] is None or item['y0'] < self.chars_box[1]:
                self.chars_box[1] = item['y0']
            if self.chars_box[2] is None or item['x1'] > self.chars_box[2]:
                self.chars_box[2] = item['x1']
            if self.chars_box[3] is None or item['y1'] > self.chars_box[3]:
                self.chars_box[3] = item['y1']
            x, y = (item['x0'] + item['x1']) / 2, (item['y0'] + item['y1']) / 2
            if y not in chars:
                chars[y] = {}
                heights[y] = {'y0': item['y0'], 'y1': item['y1']}
                widths[y] = {'x0': item['x0'], 'x1': item['x1']}
            else:
                heights[y]['y0'] = min(heights[y]['y0'], item['y0'])
                heights[y]['y1'] = max(heights[y]['y1'], item['y1'])
                widths[y]['x0'] = min(widths[y]['x0'], item['x0'])
                widths[y]['x1'] = max(widths[y]['x1'], item['x1'])

            if x not in chars[y]:
                chars[y][x] = i

        for y in sorted(chars, reverse=True):
            if not self.ys:
                self.ys.append(y)
                self.chars[y] = chars[y]
                continue

            y_ = self.ys[-1]
            min_range = min(
                heights[y_]['y1'] - heights[y_]['y0'],
                heights[y]['y1'] - heights[y]['y0']
            )
            if heights[y_]['y1'] - heights[y]['y0'] < min_range / 2 \
                    or heights[y]['y1'] - heights[y_]['y0'] < min_range / 2:
                self.ys.append(y)
                self.chars[y] = chars[y]
            else:
                self.chars[y_].update(chars[y])
                heights[y_]['y0'] = min(heights[y]['y0'], heights[y_]['y0'])
                heights[y_]['y1'] = max(heights[y]['y1'], heights[y_]['y1'])
                widths[y_]['x0'] = min(widths[y]['x0'], widths[y_]['x0'])
                widths[y_]['x1'] = max(widths[y]['x1'], widths[y_]['x1'])
        for y in reversed(self.ys):
            self.strings[y] = {
                'text': ''.join(self.chars_ids[self.chars[y][x]]['char'] for x in sorted(self.chars[y])),
                'bounds': [widths[y]['x0'], heights[y]['y0'], widths[y]['x1'], heights[y]['y1']]
            }

    def read_vertical_borders(self):
        for (y0, y1), vertical_lines in self.vertical_lines_groups.items():
            ys = []
            for y in reversed(self.ys):
                if y < y0:
                    continue
                if y > y1:
                    break
                ys.append(y)

            if ys:
                for y in ys:
                    del self.strings[y]
                self.add_vertical_borders(vertical_lines)
                horizontal_lines = self.read_horizontal_borders(vertical_lines)
                self.add_horizontal_borders(horizontal_lines)

    def read_horizontal_borders(self, vertical_lines):
        horizontal_lines = {}

        y0s, y1s = [], []
        for x, lines in vertical_lines.items():
            for y0, y1 in lines:
                y0s.append(y0)
                y1s.append(y1)
        if not y0s:
            return horizontal_lines

        y0, y1 = min(y0s), max(y1s)
        for y in self.horizontal_lines:
            if y0 <= y <= y1 or self.near_enough(y, y0) or self.near_enough(y, y1):
                horizontal_lines[y] = self.horizontal_lines[y]

        return horizontal_lines

    def add_vertical_borders(self, vertical_lines):
        vertical_borders = {}
        y0, y1 = None, None
        for x, lines in vertical_lines.items():
            if not lines:
                continue
            y0_ = min(ly0 for ly0, ly1 in lines)
            y1_ = max(ly1 for ly0, ly1 in lines)
            if y0 is None or y0 > y0_:
                y0 = y0_
            if y1 is None or y1 < y1_:
                y1 = y1_
            if x not in self.vertical_borders:
                vertical_borders[x] = lines
            else:
                vertical_borders[x].extend(lines)
        self.vertical_borders[y0, y1] = vertical_borders

    def add_horizontal_borders(self, horizontal_lines):
        for y, lines in horizontal_lines.items():
            if y not in self.horizontal_borders:
                self.horizontal_borders[y] = lines
            else:
                self.horizontal_borders[y].extend(lines)

    def read_tables(self):
        horizontal_border_ys = sorted(self.horizontal_borders.keys())
        for (b, e), vertical_borders in self.vertical_borders.items():
            table = {
                'bounds': [None, b, None, e],
                'horizontal_borders': {},
                'vertical_borders': vertical_borders,
                'cells': {},
                'width': 0,
                'height': 0,
            }
            for i, y in enumerate(horizontal_border_ys):
                if b > y and not self.near_enough(b, y):
                    continue
                if b <= y <= e or self.near_enough(b, y) or self.near_enough(e, y):
                    table['horizontal_borders'][y] = self.horizontal_borders[y]
                    for x0, x1 in self.horizontal_borders[y]:
                        if table['bounds'][0] is None or x0 < table['bounds'][0]:
                            table['bounds'][0] = x0
                        if table['bounds'][2] is None or x1 > table['bounds'][2]:
                            table['bounds'][2] = x1
                else:
                    horizontal_border_ys = horizontal_border_ys[i:]
                    break

            if not table['vertical_borders'] or not table['horizontal_borders']:
                continue

            text_ids = self.crop([table['bounds'][0], table['bounds'][1], min(table['vertical_borders']), table['bounds'][3]])
            if any(self.chars_ids[i]['char'].strip() for i in text_ids):
                table['vertical_borders'][table['bounds'][0]] = [(table['bounds'][1], table['bounds'][3])]
            else:
                table['vertical_borders'][min(table['vertical_borders'])] = [(table['bounds'][1], table['bounds'][3])]

            text_ids = self.crop([max(table['vertical_borders']), table['bounds'][1], table['bounds'][2], table['bounds'][3]])
            if any(self.chars_ids[i]['char'].strip() for i in text_ids):
                table['vertical_borders'][table['bounds'][2]] = [(table['bounds'][1], table['bounds'][3])]
            else:
                table['vertical_borders'][max(table['vertical_borders'])] = [(table['bounds'][1], table['bounds'][3])]

            text_ids = self.crop([table['bounds'][0], table['bounds'][1], table['bounds'][2], min(table['horizontal_borders'])])
            if any(self.chars_ids[i]['char'].strip() for i in text_ids):
                table['horizontal_borders'][table['bounds'][1]] = [(table['bounds'][0], table['bounds'][2])]
            else:
                table['horizontal_borders'][min(table['horizontal_borders'])] = [(table['bounds'][0], table['bounds'][2])]

            text_ids = self.crop([table['bounds'][0], max(table['horizontal_borders']), table['bounds'][2], table['bounds'][3]])
            if any(self.chars_ids[i]['char'].strip() for i in text_ids):
                table['horizontal_borders'][table['bounds'][3]] = [(table['bounds'][0], table['bounds'][2])]
            else:
                table['horizontal_borders'][max(table['horizontal_borders'])] = [(table['bounds'][0], table['bounds'][2])]

            self.read_cells(table)
            if not table['cells']:
                continue
            table['width'] = max(i[1] for i in table['cells']) + 1
            table['height'] = max(i[0] for i in table['cells']) + 1
            self.tables.append(table)

    def read_cells(self, table):
        cells = table['cells']
        horizontal_borders = table['horizontal_borders']
        vertical_borders = table['vertical_borders']
        x0, y1 = None, None
        for i, y in enumerate(sorted(horizontal_borders.keys(), reverse=True)):
            if i == 0:
                y1 = y
                continue
            for j, x in enumerate(sorted(vertical_borders.keys())):
                if j == 0:
                    x0 = x
                    continue

                left_merged = self.is_merged((y1 + y) / 2, vertical_borders[x0])
                up_merged = self.is_merged((x0 + x) / 2, horizontal_borders[y1])

                current_cell_i = i - 1
                current_cell_j = j - 1
                cells[(current_cell_i, current_cell_j)] = {
                    'bounds': [x0, y, x, y1],
                    'text': '',
                    'left_merged': left_merged,
                    'up_merged': up_merged,
                }
                while left_merged or up_merged:
                    if left_merged:
                        if not current_cell_j:
                            break
                        current_cell_j -= 1
                        cells[(current_cell_i, current_cell_j)]['bounds'][2] = x
                        cells[(current_cell_i, current_cell_j)]['bounds'][1] = y
                    elif up_merged:
                        if not current_cell_i:
                            break
                        current_cell_i -= 1
                        cells[(current_cell_i, current_cell_j)]['bounds'][2] = x
                        cells[(current_cell_i, current_cell_j)]['bounds'][1] = y
                    left_merged = cells[(current_cell_i, current_cell_j)]['left_merged']
                    up_merged = cells[(current_cell_i, current_cell_j)]['up_merged']

                x0 = x
            y1 = y

        for cell in table['cells'].values():
            cell['text'] = ''
            for text in [self.chars_ids[i] for i in self.crop(cell['bounds'])]:
                cell['text'] += text['char']

    def is_merged(self, p, lines):
        for b, e in lines:
            if b <= p <= e or self.near_enough(p, b) or self.near_enough(e, p):
                return False
        return True

    def crop(self, bounds):
        text_ids = []
        x0, y0, x1, y1 = bounds
        for y in self.ys:
            if y + self.THRESHOLD < y0:
                continue
            if y > y1 + self.THRESHOLD:
                continue
            for x in sorted(self.chars[y]):
                text = self.chars_ids[self.chars[y][x]]
                x0_, y0_, x1_, y1_ = text['x0'], text['y0'], text['x1'], text['y1']
                tx, ty = (x0_ + x1_) / 2, (y0_ + y1_) / 2
                if y0 < ty < y1 and x0 < tx < x1:
                    text_ids.append(self.chars[y][x])
        return text_ids



from setuptools import setup

README = open('README.md', 'r').read()

setup(
    name='naivepdf',
    version='0.3',
    author='naivefeeling',
    author_email='625424539@qq.com',
    long_description=README,
    long_description_content_type="text/markdown",
    packages=['naivepdf', 'naivepdf.decoder', 'naivepdf.encoding', 'naivepdf.utils'],
    package_data={'naivepdf': ['cmap/*/*', 'Adobe-Core35_AFMs-314/*']},
    description='yet another pdf texts and tables extractor',
    url='https://github.com/naivefeeling/naivepdf.git',  # use the URL to the github repo
    keywords=['pdf', 'pdfparser', 'pdfextract'],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
    ],
)



from io import BytesIO


class CorruptDataException(Exception):
    pass


class LZWDecoder(object):
    def __init__(self, fp=None):
        self.fp = fp
        self.next_read_num = 9
        self.table = []
        self.prev = b''
        self._buffer = None
        self._buffer_remain = 0

    @property
    def buffer(self):
        if self._buffer_remain == 0:
            tmp = self.fp.read(1)
            if not tmp:
                raise EOFError()
            self._buffer = ord(tmp)
            self._buffer_remain = 8
        return self._buffer

    def read(self):
        num = self.next_read_num
        code = 0
        while num:
            buffer = self.buffer
            if self._buffer_remain < num:
                num -= self._buffer_remain
                code = (code << self._buffer_remain) | (buffer & ((1 << self._buffer_remain) - 1))
                self._buffer_remain = 0
            else:
                self._buffer_remain -= num
                code = (code << num) | ((buffer >> self._buffer_remain) & ((1 << num) - 1))
                break
        return code

    def feed(self, code):
        data = b''
        if code == 256:
            self.table = [bytes([c]) for c in range(256)]  # 0-255
            self.table.append(None)  # 256
            self.table.append(None)  # 257
            self.prev = b''
            self.next_read_num = 9
        elif code == 257:
            pass
        elif not self.prev:
            data = self.table[code]
            self.prev = data
        else:
            if code < len(self.table):
                data = self.table[code]
                self.table.append(self.prev + data[:1])
            elif code == len(self.table):
                self.table.append(self.prev + self.prev[:1])
                data = self.table[code]
            else:
                raise CorruptDataException
            table_length = len(self.table)
            if table_length == 511:
                self.next_read_num = 10
            elif table_length == 1023:
                self.next_read_num = 11
            elif table_length == 2047:
                self.next_read_num = 12
            self.prev = data
        return data

    def decode(self):
        while True:
            try:
                code = self.read()
            except EOFError:
                break
            try:
                data = self.feed(code)
            except CorruptDataException:
                break
            yield data


def lzw_decode(data):
    fp = BytesIO(data)
    decoder = LZWDecoder(fp)
    return b''.join(decoder.decode())





def html(f, data):
    f.write('<meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>')
    f.write('<link href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" rel="stylesheet">')
    _html(f, data)


def _html(f, data):
    for d in data:
        if 'text' in d:
            for t in d['text'].split('\n'):
                f.write(f"<p style='text-indent:2em'>{t}</p>")
        elif 'cells' in d:
            f.write('<table border="1" cellspacing="0">')
            for i in range(d['height']):
                f.write('<tr>')
                for j in range(d['width']):
                    if d['cells'][i, j]['left_merged'] or d['cells'][i, j]['up_merged']:
                        continue
                    colspan = 1
                    while colspan < d['width'] - j and d['cells'][i, j + colspan]['left_merged']:
                        colspan += 1
                    rowspan = 1
                    while rowspan < d['height'] - i and d['cells'][i + rowspan, j]['up_merged']:
                        rowspan += 1
                    f.write(f"<td colspan='{colspan}' rowspan='{rowspan}' align='center'>{d['cells'][i, j]['text']}</td>")
                f.write('</tr>')
            f.write('</table>')




import re
import logging.config

from .latin_character_encoding import LATIN_CHARACTER_ENCODINGS
from .glyphlist import GLYPH_TO_UNICODE
from .zapfdingbats import ZAPFDINGBATS
from ..config import LOGGING

logging.config.dictConfig(LOGGING)
logger = logging.getLogger()

STRIP_NAME = re.compile(r'uni([0-9]+)')


def glyph_name_to_unicode(name):
    if name in GLYPH_TO_UNICODE:
        return GLYPH_TO_UNICODE[name]
    if name in ZAPFDINGBATS:
        return ZAPFDINGBATS[name]

    # uni2713: '\u2713'
    m = STRIP_NAME.search(name)
    if not m:
        raise KeyError(name)
    return chr(int(m.group(1), 16))


ENCODINGS = {
    b'StandardEncoding': {},
    b'MacRomanEncoding': {},
    b'WinAnsiEncoding': {},
    b'PDFDocEncoding': {},
}

for (glyph_name, std, mac, win, pdf) in LATIN_CHARACTER_ENCODINGS:
    u = glyph_name_to_unicode(glyph_name)
    if std:
        ENCODINGS[b'StandardEncoding'][std] = u
    if mac:
        ENCODINGS[b'MacRomanEncoding'][mac] = u
    if win:
        ENCODINGS[b'WinAnsiEncoding'][win] = u
    if pdf:
        ENCODINGS[b'PDFDocEncoding'][pdf] = u


def get_encoding(name, differences=None):
    cid_to_unicode = ENCODINGS.get(name, ENCODINGS[b'StandardEncoding']).copy()
    if differences:
        cid = 0
        for d in differences:
            if isinstance(d, int):
                cid = d
            else:
                try:
                    cid_to_unicode[cid] = glyph_name_to_unicode(d.decode())
                except KeyError:
                    pass
                    # logger.warning(f'no cid name: {cid}, {d.decode()}')
                    # Type3 CharProcs
                cid += 1
    return cid_to_unicode



LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(asctime)s-%(name)s-%(levelname)s-%(message)s'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'simple',
            'stream': 'ext://sys.stdout'
        },
    },
    'loggers': {
        'init': {
            'level': 'ERROR',
            'handlers': ['console'],
            'propagate': 'no'
        }
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    }
}


from io import BytesIO
from .pdfpage import PDFPage, PDFPageInterpreter
from .pdfparser import PDFParser
from .pdfobject import PDFObj, PDFIndirectRef, PDFStream
from .pdfdecryption import decryption
from .exception import PDFObjNotFoundException, PDFDocumentException, PDFSyntaxException, PDFParserException, PDFParserDoneException


class PDFDocument(object):
    PAGE_INHERITABLE_ATTRS = {b'Resources', b'MediaBox', b'CropBox', b'Rotate'}

    def __init__(self, fp, password=''):
        self.parser = PDFParser(fp)
        self.password = password
        self.decipher = None
        self._startxref = None
        self._xrefs = None
        self._root = None
        self._info = []
        self._obj_cache = {}
        self._obj_stream_cache = {}
        self._indirect_ref_cache = {}

        self.fonts = {}
        self.color_spaces = {}
        self.x_objects = {}
        self.properties = {}

    @property
    def startxref(self):
        if self._startxref is None:
            startxref = None
            for line in self.parser.lexer.reversed_lines():
                line = line.strip()
                if line == b'startxref':
                    break
                startxref = line
            else:
                raise PDFDocumentException('Startxref Not Found')
            try:
                self._startxref = int(startxref)
            except:
                raise PDFDocumentException('Startxref Not Found')
        return self._startxref

    @property
    def xrefs(self):
        if self._xrefs is None:
            self._xrefs = []
            self.xref(self.startxref)
        return self._xrefs

    def xref(self, startxref):
        self.parser.seek(startxref)
        xref = self.parser.next()
        self._xrefs.append(xref)

        if b'Encrypt' in xref.trailer:
            ids = self.resolve_indirect_ref(xref.trailer[b'ID'])
            encrypt = self.resolve_indirect_ref(xref.trailer[b'Encrypt'])
            self.decipher = decryption[encrypt[b'V'].value](ids, encrypt, self.password)
        if b'Root' in xref.trailer:
            self._root = xref.trailer[b'Root']
        if b'Info' in xref.trailer:
            self._info.append(xref.trailer[b'Info'])

        if b'XRefStm' in xref.trailer:
            self.xref(int(xref.trailer[b'XRefStm']))
        if b'Prev' in xref.trailer:
            self.xref(int(xref.trailer[b'Prev']))

    @property
    def catalog(self):
        if self._root is None:
            _ = self.xrefs
            self._root = self.resolve_indirect_ref(self._root)
            self._root[b'Pages'] = self.resolve_indirect_ref(self._root[b'Pages'])
        return self._root

    @property
    def pages(self):
        for page in self._pages(self.catalog[b'Pages']):
            i = PDFPageInterpreter(page)
            i.interpret()
            yield {
                'box': [i.value for i in page.crop_box],
                'rotate': page.rotate,
                'data': i.data,
            }

    def _pages(self, page_tree):
        for kid in self.resolve_indirect_ref(page_tree[b'Kids']):
            kid = self.resolve_indirect_ref(kid)
            for k in self.PAGE_INHERITABLE_ATTRS:
                if k not in kid:
                    kid[k] = page_tree.get(k)
            if kid[b'Type'] == b'Page':
                yield PDFPage(self, kid)
            elif kid[b'Type'] == b'Pages':
                yield from self._pages(kid)

    def resolve_indirect_ref(self, o):
        while isinstance(o, PDFIndirectRef):
            key = (o.number, o.generation)
            if key not in self._indirect_ref_cache:
                o = self.get_obj(o.number)
                if isinstance(o, PDFObj):
                    o = o.value
                self._indirect_ref_cache[key] = o
            o = self._indirect_ref_cache[key]
        return o

    def deep_resolve_indirect_ref(self, o):
        o = self.resolve_indirect_ref(o)
        if isinstance(o, list):
            return [self.deep_resolve_indirect_ref(i) for i in o]
        if isinstance(o, dict):
            return {k: self.deep_resolve_indirect_ref(v) for k, v in o.items()}
        return self.resolve_indirect_ref(o)

    def decrypt(self, number, generation, value):
        if isinstance(value, list):
            value = [self.decrypt(number, generation, v) for v in value]
        elif isinstance(value, dict):
            value = {k: self.decrypt(number, generation, v) for k, v in value.items()}
        elif isinstance(value, bytes):
            value = self.decipher.decrypt(number, generation, value)
        elif isinstance(value, PDFStream):
            value.v = self.decipher.decrypt(number, generation, value.v, value.dict)
        return value

    def get_obj(self, obj_number):
        if obj_number not in self._obj_cache:
            for xref in self.xrefs:
                try:
                    stream_obj_number, position, gen_no = xref.obj_position(obj_number)
                except PDFObjNotFoundException:
                    continue
                try:
                    if stream_obj_number is None:
                        obj = self.obj(position, obj_number)
                        if self.decipher:
                            obj.value = self.decrypt(obj.number, obj.generation, obj.value)
                    else:
                        obj = self.obj_in_stream(position, stream_obj_number)
                    break
                except (PDFParserException, PDFSyntaxException):
                    continue
            else:
                raise PDFObjNotFoundException(obj_number)
            self._obj_cache[obj_number] = (obj, gen_no)
        obj, gen_no = self._obj_cache[obj_number]
        return obj

    def obj(self, position, obj_number):
        self.parser.seek(position)
        obj = self.parser.next()
        if not isinstance(obj, PDFObj):
            raise PDFSyntaxException(f'Not A PDFObj: {obj}')
        if obj.number != obj_number:
            raise PDFSyntaxException(f'Obj Mismatch: {obj.number} != {obj_number}')
        if isinstance(obj.value, PDFStream):
            obj.value.dict = self.deep_resolve_indirect_ref(obj.value.dict)
        return obj

    def obj_in_stream(self, position, stream_obj_number):
        stream_obj = self.get_obj(stream_obj_number)
        if stream_obj_number not in self._obj_stream_cache:
            stream = stream_obj.value
            parser = PDFParser(BytesIO(stream.value))
            objs = []
            while True:
                try:
                    o = parser.next()
                except PDFParserDoneException:
                    break
                objs.append(o)
            n = stream.dict[b'N'].value if b'N' in stream.dict else 0
            self._obj_stream_cache[stream_obj_number] = (n, objs)
        n, objs = self._obj_stream_cache[stream_obj_number]
        obj = objs[n * 2 + position]
        return obj



import os
from .encoding.encoding import glyph_name_to_unicode


class AFMLoader(object):
    def __init__(self):
        self.fonts = {}
        self.load_afm()

    def load_afm(self):
        afm_path = os.path.join(os.path.dirname(__file__), 'Adobe-Core35_AFMs-314')
        for d in os.listdir(afm_path):
            if not d.endswith('.afm'):
                continue
            font = {
                'descriptor': {},
                'metrics': {},
            }
            with open(os.path.join(afm_path, d)) as f:
                for line in f:
                    line = line.strip().replace('\t', ' ')
                    if not line:
                        continue
                    row = line.split(' ')

                    k1 = {
                        'FontName': b'FontName',
                        'FamilyName': b'FontFamily',
                        'Weight': b'FontFamily'
                    }
                    k2 = {
                        'Ascender': b'Ascent',
                        'Descender': b'Descent',
                        'CapHeight': b'CapHeight',
                        'XHeight': b'XHeight',
                        'ItalicAngle': b'ItalicAngle',
                    }

                    if row[0] in k1:
                        font['descriptor'][k1[row[0]]] = row[1]
                    elif row[0] in k2:
                        font['descriptor'][k2[row[0]]] = float(row[1])
                    elif row[0] == 'IsFixedPitch' and row[1].lower() == 'true':
                        font['descriptor'][b'Flags'] = 64
                    elif row[0] == 'FontBBox':
                        font['descriptor'][b'FontBBox'] = tuple(map(float, row[1:5]))
                    elif row[0] == 'C':
                        row = line.split(';')
                        char = None
                        metrics = {}
                        for l in row:
                            r = l.strip().split(' ')
                            if r[0] == 'N':
                                char = glyph_name_to_unicode(r[1])
                            elif r[0] in {'C', 'WX'}:
                                metrics[r[0]] = int(r[1])
                            elif r[0] == 'B':
                                metrics['B'] = tuple(map(float, r[1:5]))
                        if char is not None:
                            font['metrics'][char] = metrics
            if font['descriptor'].get(b'FontName'):
                self.fonts[font['descriptor'][b'FontName']] = font


afm_loader = AFMLoader()